# The module provides an OO  interface to DS.
# The module is desigened to be used in a DNSSEC registry.
# perldoc DNSSEC::Registry::TRUST_DB;
#


package DNSSEC::Registry::Trust_DB;
use DNSSEC::Registry::Trust_Container;
use IO::File;
use DB_File::Lock;
use Data::Dumper;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK   $trust_db_err);




require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

@EXPORT = qw(
	     $trust_db_err	
);

$VERSION = (qw$Revision: 11 $)[1];




my $debug=0;


my $TRUST_PRODUCTION_ROOT   = "/home/noc/TrustDB/";
$TRUST_PRODUCTION_ROOT=$ENV{TRUST_PRODUCTION_ROOT} if $ENV{TRUST_PRODUCTION_ROOT};



#
#  Some programming and/or maintenance notes
#
#

#  The DB is based on a tied hash. Locking takes place at the moment of tying.
#  unlocking takes place at the moment of untying.
#  

#  The db object is not providing the tie itself but stores some properties of
#  the Berkeley db like the file name.
#


#  Any object that uses the database will individually lock the
#  database at read or write

#  Multiple trustobject are stored under the same hash-key (the domain
#  name) There may be multiple keys for the same domain, even keys
#  with the same domainname, algorithm and key-id. The are stored as
#  'strings' use the _value_to_objects and the _objects_to_value to
#  store or retrieve an array of trust objects from the hash.  # All
#  this is ugly and can be improved... Go ahead and do it. All
#  methods # starting with an underscore should not be used in
#  applications, all the # methods documented (perldoc) will need to
#  be ported if you change the DB backend which is, at this time,
#  quite primitive.



=head1 NAME

DNSSEC::Registry::TRUST_DB - Perl extension the implementation of a
trust database.

=head1 SYNOPSIS

  use DNSSEC::Registry::TRUST_DB;



=head1 DESCRIPTION

The DNSSEC::Registry::TRUST_DB is an abstract interface to a database that
contains Trust_Contianer objects.

For now only Trust_Containers that enclose a DNSKEY RR can be used.




=head1 METHODS

=head2 new

    $trustdb=DNSSEC::Registry::TRUST_DB->new($parent);

Provides an object with access to the database for zone with apex
$parent. 

The new method will do some rudimentary checking on permissions. It
returns 0 on failure.


Unforutunattely the prototype does not take "." as an argument so you
cannot play 'root'.

=cut


sub new {
    my $class=shift;
    my $domain=shift;
    $domain =~ s/\.$//;
    
    $trust_db_err="No Error";
    if (! -d $TRUST_PRODUCTION_ROOT ){
	$trust_db_err="Could not open $TRUST_PRODUCTION_ROOT";
	return (0);
    }
    $TRUST_PRODUCTION_ROOT=~ s!/$!! ;
    
    if (! defined($domain) ){
	$trust_db_err="Undefined domain";
	return (0);
    }
    my $self={
	_domain => $domain,
    };

    bless $self,  $class;
    return $self;
}







=head2 store

    my trustobj=NSSEC::Registry::Trust_Container($keyrr);
    $trustdb->store($trustobj);

Stores the $trustobj in the database overwriting possible existing trustobj
with the same keyrr.

One should test for the existence (using in_db) of a $keyrr in the DB
for using the store method if one does not want to overwrite original
vallues

returns 0 on failure and setst $trust_db_err;

=cut 


sub store {
    # Returns 1 on if objects exists
    my $self=shift;
    my $trustobj=shift;

    print "Store called\n" if $debug;
    
    return 0 if  $self->_incompatible($trustobj);
    my $keyname=$self->_db_key_name($trustobj->RR);
    my %db;
    $self->tiehash(\%db,$self->lock("write"));
    if (! exists $db{$keyname} ){
	$db{$keyname}=$self->_objecst_to_value($trustobj);
	return 1;
    }
    my @objects_in_db= $self->_value_to_objects($db{$keyname});

    my @new_trust_objects;
    
    # Copy all the trust object to the new array, replacing an eventual
    # old one

    # Here we also need to check if a DNSKEY with the same
    # domain-algorithm-keyid triplet exists than it will need to be
    # revoked.
    
    my $is_replaced=0;
    my $object_now_in_db;
    my %triplet_test;
    my $i=0;
    $trust_db_err="No Error";
    foreach $object_now_in_db ( @objects_in_db ){
	
	if ($object_now_in_db->identifyer eq $trustobj->identifyer ) {

	    push  @new_trust_objects, $trustobj;
	    $is_replaced=1;
	}else{
	    # Go out of the loop if a object is found with the same
	    # domain-alg-keyid that is not revoked. (We have two keys
	    # with the same key-tag but different keymaterial)
	    if ( $trustobj->RR->keytag == $object_now_in_db->RR->keytag &&
		 $trustobj->RR->algorithm == $object_now_in_db->RR->algorithm &&
	          ! $object_now_in_db->is_revoked ){
		
		$trust_db_err="Non revoked key with same domain key_id and algorithm in DB";
		return 0;
	    }
	    
	    
	    push  @new_trust_objects, $object_now_in_db;
	}
    }
    push  @new_trust_objects, $trustobj unless $is_replaced;
    $db{$keyname}=$self->_objecst_to_value(@new_trust_objects);
    untie(%db);
    return 1;
}


=head2 in_db

    my $trustobj=NSSEC::Registry::Trust_Container->new($keyrr);
    $trustdb->in_db($trustobj);


Returns true if the a $trustobj containing the same $key RR is
allready in the database. Returns 0 if the object is not in the DB or
if an error occured. More information on the error is in
$trust_db_err.

=cut

sub in_db {
    # Returns 1 on success
    my $self=shift;
    my $trustobj=shift;
    $trust_db_err="No Error";
    if ($self->_incompatible($trustobj)){
	$trust_db_err= "Incompatible argument (".$trust_db_err.")" ;
	return 0;
    }
    my $db_key=$self->_db_key_name($trustobj->RR);
    my %db;
    $self->tiehash(\%db,$self->lock("write"));
    if ( exists($db{$db_key})){
	# There may be multiple keys with the same keyid. 
	foreach my $tr_o (   $self->_value_to_objects( $db{$db_key})){
	    if ($tr_o->identifyer eq $trustobj->identifyer) {
		untie %db;
		return 1;
	    }
	}    
    };
    untie(%db);
    return 0;
    
}



=head2 get

    my @trustcontainers=$trustdb->dump_db($domain,<$algorithm,<$key-id>>);

Returns an array of Trust_Container objects that match the
$domain. The query may be futher specified by specifying both the
algorithm number or the algorithm number and the key-id.

Returns 0 on error and sets $trust_db_err.  $trust_db_err is set to
"No Match" if no data is found for the query.

It is possible that for the a domain, algorithm, key-id truplet
multiple keys exist in the database; the key material itself is the
real unique key in the DB.


=cut


sub get {
    my $self=shift;
    my $domain=shift;
    my $algorithm=shift;
    my $keytag=shift;



#
# TODO handle the case when algorithm is not defined and keytag is...
#      parameter handling needs to be updated just a tiny weeny bit here.

    
    $trust_db_err="No Match";
    my %db;
    my @trustobjects;
    $self->tiehash(\%db,$self->lock("read"));
    my @returnarray;
    
    if (defined $domain){
	if (exists $db{$domain}){
	    @trustobjects=$self->_value_to_objects( $db{$domain} );
	}
    }else{  
	# domain is not defined;
	# Return all the keyobjects in the database.
	foreach my $db_key (keys %db){
	    push @returnarray, $self->_value_to_objects( $db{$db_key})
	    }
	untie(%db);
	return @returnarray;
    }
    untie(%db);
    
    # A domain was defined as argument @trustobjects contains zero or more
    # trust objects. We may need to select more specifics.

    foreach  my $trustobj (@trustobjects) {
	print "Get works on\n".$trustobj->string."\n" if $debug;
	if (defined $keytag){
	    if (defined $algorithm){
		# keytag and algorithm defined
		push @returnarray, $trustobj if 
		    ( $algorithm==$trustobj->RR->algorithm &&  
		      $keytag==$trustobj->RR->keytag);
	    }else{
		# only keytag defined
		push @returnarray, $trustobj if $keytag==$trustobj->RR->keytag;
	    }
	}elsif(defined $algorithm){
	    #only algorithm defined
	    push @returnarray ,$trustobj if $algorithm==$trustobj->RR->algorithm ;
	}else{
	    # both $keytag and %algorithm not defined
	    push @returnarray, $trustobj;
	}
    }
    return @returnarray;
    
}


=head2 get_trusted, get_pending, get_expirerd
   
    @trusted_objects=$trustdb->get_trusted(foo.examle, 5, 12311);
    @pending_objects=$trustdb->get_pending(foo.example);
    @expired_objects=$trustdb->get_expired();


Returns an array of Trust_container objecst which state at t=now is
trusted, pending or expired respecively. The argument refined to the
search to a $domain, optionally  the $keytag and/or the 
$algorithm can be supplied together with the domain.




=cut


# These get_ functions essentially do the same .. maybe they should be
# merged into one function

sub get_trusted{
    my $self=shift;
    my @trustobjects=$self->get(@_);
    my $trustobj;

    my @trusted_trustobjects;

    foreach  $trustobj (@trustobjects) {
	push  @trusted_trustobjects , $trustobj if $trustobj->is_trusted;
    }
    return @trusted_trustobjects;
}





sub get_pending {
    my $self=shift;
    my @trustobjects=$self->get(@_);

    print "get_pending called\n" if $debug;

    my @trusted_trustobjects;
    my $trustobj;

    foreach $trustobj (@trustobjects) {
	push  @trusted_trustobjects , $trustobj if $trustobj->is_pending;
    }

    return @trusted_trustobjects;
}




sub get_expired {
    my $self=shift;
    my @trustobjects=$self->get(@_);
    print "get_expired called\n" if $debug;

    my @trusted_trustobjects;
    my $trustobj;

    foreach $trustobj (@trustobjects) {
	push  @trusted_trustobjects , $trustobj if $trustobj->is_expired;
    }

    return @trusted_trustobjects;
}






sub get_revoked {
    my $self=shift;
    my @trustobjects=$self->get(@_);

    print "get_revoked called\n" if $debug;

    my @trusted_trustobjects;
    my $trustobj;

    foreach $trustobj (@trustobjects) {
	push  @trusted_trustobjects , $trustobj if $trustobj->is_revoked;
    }

    return @trusted_trustobjects;
}



=head2 dump_db

    @trustobjects=$trustdb->dump_db;
    @trustobjects=$trustdb->dump_db(1);

Dumps the whole content of the database.  If the argument is true 
then objects are printed to STDOUT.

=cut



sub dump_db{
    my $self=shift;
    my $printthem=shift;
    my @trustobjects=$self->get();
    my $trustobj;

    print "Dump_DB called\n" if $debug;
    foreach  $trustobj (@trustobjects){
	$trustobj->print if $printthem;
	
    }
    
    return @trustobjects ;
    
}


=head2 any_trusted 
    if ($self->any_trusted($keyset));

This method takes a Net::DNS::Keyset object as argument and returns
true if the keyset is consistent and signed by at least one key that
is_trusted.

Returns 0 if no key is false or if an error occcured. If an error
occured $trust_db_err is set. If no error occured $trust_db_err equals
"No Error".

=cut


sub any_trusted {
    my $self=shift;
    my $keyset=shift;
    $trust_db_err="No Error";
    if ( ! $keyset->isa("Net::DNS::Keyset") ){
	return  $trust_db_err="Comme on, the argument is not a keyset" ;
	};
    return 0 unless $keyset->verify;
    foreach  my $key ( $keyset->keys ){
	my @trust_objects=$self->get($key->name,$key->algorithm,$key->keytag);
	foreach my $trust_object (@trust_objects){
	    return 1 if ( $trust_object->RR->bin eq $key->bin ) && 
		$trust_object->is_trusted;
	}
    }
    return 0;
}
    
#######################################################################
#
#  Helper classes for internal usage
#  No perlpod documentation.
#
#######################################################################


sub _incompatible {
    # Returns a 0 value if all is correct and returns and sets
    # $trust_db_err if the argument cannot be used in combination with
    # the DB for some reason
    $trust_db_err="No Error";
    my $self=shift;
    my $trustobj=shift;
    my $key=$trustobj->RR;
    return $trust_db_err="Enclosed RR is not a key " if ($key->type ne "DNSKEY");
    my $domain=$self->domain;
    $domain=~s/\./\\\./g;
    return  $trust_db_err="The trust container is not in the $domain namespace" if ($key->name !~ /\.$domain$/);
    return 0;
}


#  Returns the domain attribute.
#
sub domain {
    my $self=shift;
    return $self->{"_domain"};
}





#
#  generates a key given a $key rr object as input.
#
sub _db_key_name{
    # The keyname is just the name of the domain.
    # More details need to be obtained from within the object
       my $self=shift;
    my $key=shift;
    return $key->name;
}



# $self->lock("read");
#
#  Returns a reference to a lock hash. Takes "read" or "write" as input.

sub lock {
    my $self=shift;
    my $lock_mode=shift;
    die "lock_mode should be read or write ($lock_mode)" if ($lock_mode ne "read" && $lock_mode ne "write");
    my $domain=$self->domain;

    return {
	mode            => $lock_mode,
	nonblocking     => 0,
	lockfile_name   => $TRUST_PRODUCTION_ROOT."/". "$domain.shared.lock",
	lockfile_mode   => 0666,
    };
    
}

# $self->tiehash(\%db,$self->lock("read");
#
# Tiehash takes to arguments. A reference to a hash that needs to be tied and
# a reference to a lock hash. The lock hash should be generated with the
# lock method above.

sub tiehash {
    my $self=shift;
    my $trustdb=shift;
    my $lock=shift;
    my $domain=$self->domain;
    my $filename=$TRUST_PRODUCTION_ROOT."/".$domain.".db";
    my $DB;
    $DB=tie %{$trustdb}, 'DB_File::Lock', 
    $filename, O_RDWR|O_CREAT, 0666, $DB_HASH , $lock
	or die "Cannot open  : " .$filename."( $! )\n" ;
    # This is uggly.. but it seems that the tie does not set the 
    # file creation mode correctly.
    chmod(0666, $filename);
}	





#  The delimiter used for internal storage
#
#  These two functions turn a bunch of objects into a string that can 
#  be stored as a vallue in the db.
#  Perl is a little troublesome if it commes to storing weird datastructures.

my $value_delimiter="\n-----------Delimiter KEYDB version 0.1--------------\n";

sub _value_to_objects {
    my $self=shift;
    my $instring=shift;

    my @trust_objects;
    my @object_strings=split /$value_delimiter/ ,$instring;

    my $object_string;
    foreach $object_string (@object_strings){
	my $trust_object=DNSSEC::Registry::Trust_Container->from_string($object_string);
	push @trust_objects, $trust_object;
	}
    return @trust_objects;
}
	
sub _objecst_to_value {
    my $self=shift;
    my @trust_objects=@_;
    my $trust_object;
    my $outstring;
    foreach $trust_object (@trust_objects) {
	$outstring.=$trust_object->string;
	$outstring.=$value_delimiter;
    }
    $outstring=~ s/$value_delimiter$//; # strip last 
    return $outstring;
}




#
#   This removes the database file... The function is only there for the 
#   test suite to run cleanly....
#
#   Use with care!!!!!!!

sub _Delete_DB {
    my $self=shift;
    my $domain=$self->domain;
    my $filename=$TRUST_PRODUCTION_ROOT."/".$domain.".db";    
    unlink($filename);
    return 1;

}


=head1 DB DESIGN


For developers mostly.

The Database is based on hash tied to a Berkely DB. The keys to the
hash are the dnames of the keys stored. The vallues are arrays of
Trust_Container objects.


=head1 COPYRIGHT

Copyright (c) 2002 RIPE NCC.  Author Olaf M. Kolkman
<net-dns-sec@ripe.net>

All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO
EVENT SHALL AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.


=head1 SEE ALSO

perl(1). Net::DNS, Net::DNS::SEC

=cut





1;
__END__



